"""
Parameter handling with YAML
"""

import os
import traceback
import yaml

import numpy as np
from .. import prep, utils

# ACS/WFC & WFC3 UVIS + IR filters
IR_N_FILTERS = ['F126N', 'F128N', 'F130N', 'F132N', 'F164N', 'F167N']
IR_M_FILTERS = ['F098M', 'F127M', 'F139M', 'F153M']
IR_W_FILTERS = ['F105W', 'F110W', 'F125W', 'F140W', 'F160W']

NIRISS_FILTERS = ['F090W', 'F115W', 'F150W', 'F200W', 
                  'F277W', 'F356W', 'F444W',
                  'F140M', 'F158M', 'F380M', 'F430M', 'F480M']
                   
NIRCAM_LW_FILTERS = ['F322W2', 'F277W', 'F356W', 'F444W', 
                    'F250M', 'F300M', 'F335M', 'F360M', 'F410M',
                    'F430M', 'F460M', 'F480M',
                    'F323N', 'F405N', 'F466N', 'F470N']

NIRCAM_SW_FILTERS = ['F150W2', 'F070W', 'F090W', 'F115W', 'F150W', 'F200W', 
                     'F140M', 'F162M', 'F182M', 'F210M', 
                     'F164N', 'F187N', 'F212N']

MIRI_FILTERS = ['F560W', 'F770W', 'F1000W', 'F1130W', 'F1280W', 
                'F1500W', 'F1800W', 'F2100W', 'F2550W']

JW_FILTERS = np.unique(NIRISS_FILTERS + NIRCAM_LW_FILTERS + NIRCAM_SW_FILTERS + MIRI_FILTERS).tolist()

OPT_N_FILTERS =  ['F469N', 'F487N', 'F502N']
OPT_N_FILTERS += ['FQ436N', 'FQ437N', 'FQ492N', 'FQ508N', 'FQ575N']
OPT_N_FILTERS += ['F631N', 'F645N', 'F656N', 'F657N', 'F658N', 
                  'F660N', 'F665N', 'F673N', 'F680N']
OPT_N_FILTERS += ['FQ619N', 'FQ634N', 'FQ672N', 'FQ674N', 'FQ727N', 'FQ750N', 
                  'FQ889N', 'FQ906N', 'FQ924N', 'FQ937N', 'F892N', 'F953N']

OPT_M_FILTERS = ['F410M', 'F467M', 'F547M', 'F550M', 'F621M',
                 'F689M', 'F763M', 'F845M']
OPT_W_FILTERS = ['F200LP', 'F350LP', 'F435W', 'F438W', 'F439W',
                 'F450W', 'F475W', 'F475X', 
                 'F555W', 'F569W', 'F600LP', 'F606W', 'F622W',
                 'F625W', 'F675W', 'F702W', 
                 'F775W', 'F791W', 'F814W', 'F850LP']

UV_N_FILTERS = ['F280N', 'F343N', 'F373N', 'F395N']
UV_N_FILTERS += ['FQ232N', 'FQ243N', 'FQ378N']
UV_M_FILTERS = ['F390M']
UV_W_FILTERS = ['F275W', 'F336W', 'F390W']

IR_GRISMS = ['G102', 'G141']
OPT_GRISMS = ['G800L']
UV_GRISMS = ['G280']

JW_GRISMS = ['GR150C', 'GR150R', 'GRISMR', 'GRISMC']

VALID_FILTERS = OPT_M_FILTERS + OPT_W_FILTERS + OPT_GRISMS  # + UV_GRISMS
VALID_FILTERS += IR_M_FILTERS + IR_W_FILTERS + IR_GRISMS
VALID_FILTERS += JW_FILTERS + JW_GRISMS

ALL_IMAGING_FILTERS = UV_N_FILTERS + UV_M_FILTERS + UV_W_FILTERS
ALL_IMAGING_FILTERS += OPT_N_FILTERS + OPT_M_FILTERS + OPT_W_FILTERS
ALL_IMAGING_FILTERS += IR_N_FILTERS + IR_M_FILTERS + IR_W_FILTERS

GRIS_REF_FILTERS = {'G141': ['F140W', 'F160W', 'F125W', 'F105W', 'F110W',
                             'F098M', 'F127M', 'F139M', 'F153M', 'F132N',
                             'F130N', 'F128N', 'F126N', 'F164N', 'F167N'],
                    'G102': ['F105W', 'F098M', 'F110W', 'F125W', 'F140W',
                             'F160W', 'F127M', 'F139M', 'F153M', 'F132N',
                             'F130N', 'F128N', 'F126N', 'F164N', 'F167N'],
                    'G800L': ['F814W', 'F850LP', 'F606W', 'F435W', 'F775W',
                              'F555W', 'opt'], 
                    'GR150C': ['F115W', 'F150W', 'F200W',
                               'F090W', 'CLEAR-F090W',
                               'CLEAR-F115W', 'CLEAR-F150W', 'CLEAR-F200W',
                               'F090WN-CLEAR', 'F115WN-CLEAR', 'F150WN-CLEAR',
                               'F200WN-CLEAR'
                              ],
                    'GR150R': ['F115W', 'F150W', 'F200W',
                               'F090W', 'CLEAR-F090W',
                               'CLEAR-F115W', 'CLEAR-F150W', 'CLEAR-F200W'
                               'F090WN-CLEAR', 'F115WN-CLEAR', 'F150WN-CLEAR',
                               'F200WN-CLEAR'
                               ],
                    'GRISMR': ['F277W-CLEAR','F356W-CLEAR','F410M-CLEAR',
                               'F444W-CLEAR','F277W','F356W','F410M','F444W'],
                    'GRISMC': ['F277W-CLEAR','F356W-CLEAR','F410M-CLEAR',
                               'F444W-CLEAR','F277W','F356W','F410M','F444W']
                    }


def test_aws_availability():
    """
    Test if aws s3 is available
    """
    s3_status = os.system('aws s3 ls s3://stpubdata --request-payer requester > /tmp/aws.x 2>&1')
    if s3_status == 0:
        s3_sync = 'cp'  # As of late October 2018, 's3 sync' not working with 'stpubdata'
    else:
        s3_sync = False  # Fetch from ESA archive

    return s3_sync


def write_params_to_yml(kwargs, output_file='grizli.auto_script.yml', verbose=True):
    """
    Write grizli parameters to a file
    """
    import time
    import yaml
    import grizli

    # Make copies of some parameters that can't be dumped with yaml
    try:
        phot_apertures = kwargs['multiband_catalog_args']['phot_apertures']
        kwargs['multiband_catalog_args']['phot_apertures'] = None
    except:
        phot_apertures = None

    try:
        filter_kernel = kwargs['multiband_catalog_args']['detection_params']['filter_kernel']
        kwargs['multiband_catalog_args']['detection_params']['filter_kernel'] = None
    except:
        filter_kernel = None

    # Write the file
    fp = open(output_file, 'w')
    fp.write('# {0}\n'.format(time.ctime()))
    fp.write('# Grizli version = {0}\n'.format(grizli.__version__))

    for k in kwargs:

        try:
            # Write copies of dicts
            d = {k: kwargs[k].copy()}
        except:
            # Write single variables
            d = {k: kwargs[k]}

        # flow_style = False to get the correct dict/variable formatting
        yaml.dump(d, stream=fp, default_flow_style=False)

    fp.close()

    # Revert the things we changed above
    if phot_apertures is not None:
        kwargs['multiband_catalog_args']['phot_apertures'] = phot_apertures

    if filter_kernel is not None:
        kwargs['multiband_catalog_args']['detection_params']['filter_kernel'] = filter_kernel

    if verbose:
        print('\n# Write parameters to {0}\n'.format(output_file))


def safe_yaml_loader(yamlfile, loaders=[yaml.FullLoader, yaml.Loader, None]):
    """
    Try different YAML loaders
    """
    args = None
    for loader in loaders:
        with open(yamlfile) as fp:
            try:
                args = yaml.load(fp, Loader=loader)
                break
                print(loader)
            except:
                pass

    if args is None:
        raise IOError(f'Failed to load {yamlfile} with {loaders}.')

    return args


def get_yml_parameters(local_file=None, copy_defaults=False, verbose=True, skip_unknown_parameters=True):
    """
    Read default parameters from the YAML file in `grizli/data`

    Returns:

    kwargs : dict
        Parameter dictionary (with nested sub dictionaries).

    """
    import shutil

    default_yml = 'auto_script_defaults.yml'

    path = os.path.join(os.path.dirname(__file__), '..', 'data', default_yml)

    if copy_defaults:
        if local_file is None:
            local_file = default_yml

        shutil.copy(path, local_file)
        if verbose:
            print('Copied default parameter file to {0}'.format(local_file))

        return False

    kwargs = safe_yaml_loader(path)

    if local_file is not None:

        local_args = safe_yaml_loader(local_file)

        for k in local_args:
            if (k not in kwargs) and skip_unknown_parameters:
                print('Skipping user keyword {0}'.format(k))
                continue

            # Dictionaries
            if isinstance(local_args[k], dict):
                for subk in local_args[k]:
                    if (subk not in kwargs[k]) and skip_unknown_parameters:
                        print('Skipping user keyword {0}.{1}'.format(k, subk))
                        continue

                    kwargs[k][subk] = local_args[k][subk]
            else:
                kwargs[k] = local_args[k]

    if kwargs['fetch_files_args']['s3_sync']:
        kwargs['fetch_files_args']['s3_sync'] = test_aws_availability()

    # catalog defaults
    dd = kwargs['multiband_catalog_args']
    if dd['detection_params'] is None:
        dd['detection_params'] = prep.SEP_DETECT_PARAMS

    if dd['phot_apertures'] is None:
        dd['phot_apertures'] = prep.SEXTRACTOR_PHOT_APERTURES_ARCSEC

    return kwargs


def compare_args():
    from grizli.pipeline import default_params
    y = default_params.get_yml_parameters(copy_defaults=False)

    d = default_params.get_args_dict()

    for k in y:
        if k not in d:
            print(k)
        else:
            if y[k] != d[k]:
                print('!!', k)
                print(y[k])
                print(d[k])
    pass



import unittest

import os
import glob

import numpy as np

from grizli import utils
from grizli.pipeline import auto_script

class UtilsTester(unittest.TestCase):
    
    def test_parse_wfc3_visits(self):
        """
        """
        
        # These are copies of the WFC3-ERSII demo FLT files but where 
        # the data extensions have been set to dtype=uint8 so that they 
        # gzip up small
        
        path = os.path.dirname(utils.__file__)        
        if not os.path.exists('wfc3-parse-test'):
            os.system(f'tar xzvf {path}/tests/data/wfc3-parse-test.tar.gz')
            
        os.chdir('wfc3-parse-test/Prep/')
        
        files = glob.glob('../RAW/*flt.fits')
        files.sort()
        info = utils.get_flt_info(files)
        
        assert(len(info) == len(files))
        
        _visits, _filters = utils.parse_flt_files(info=info, 
                                      uniquename=True, get_footprint=True, 
                                      use_visit=True, max_dt=0.5, 
                                      visit_split_shift=2)
        
        assert(len(_visits) == 4)
        
        _groups = utils.parse_grism_associations(_visits, info)
        assert(len(_groups) == 2)
        
        ### From auto_script
        kwargs = auto_script.get_yml_parameters(local_file='my_params.yml', 
                                                copy_defaults=False)
        
        root = 'test'
        
        for bval in [True, False]:
            kwargs['parse_visits_args']['combine_same_pa'] = bval
            _ = auto_script.parse_visits(field_root=root,
                                         **kwargs['parse_visits_args'])

            visits, groups, info = _
        
            assert(hasattr(info, 'colnames'))
            assert(len(info) == 16)
        
            assert(len(visits) == 4)
        
        assert(os.path.exists(f'{root}_visits.yaml'))
        
        __ = auto_script.load_visits_yaml(f'{root}_visits.yaml')
        _visits, _groups, _info = __
        
        for v in [visits[0], _visits[0]]:
            for k in ['product','files','footprint']:
                assert(k in v)
        
            assert(v['product'].startswith('wfc3-ers'))
            assert(len(v['files']) == 4)
        
        # Grism groups
        for g in [groups, _groups]:
            assert(len(g) == 2)
            
            gi = g[0]
            for k in ['grism', 'direct']:
                assert(k in gi)
            
            gr = gi['grism'] 
            assert(gr['product'] == 'wfc3-ersii-g01-b6o-21-119.0-g102')
            assert(len(gr['files']) == 4)
            
        os.chdir('../../')
        if os.path.exists('wfc3-parse-test'):
            os.system('rm -rf wfc3-parse-test')
            


"""
Generate PSF at an arbitrary position in a drizzled image using the WFC3/IR
effective PSFs.
"""

import os
from collections import OrderedDict

import numpy as np

import astropy.io.fits as pyfits
import astropy.wcs as pywcs

# from research.dash.epsf import DrizzlePSF

DEMO_LIST = [
    "ib6o23rtq_flt.fits",
    "ib6o23rwq_flt.fits",
    "ib6o23rzq_flt.fits",
    "ib6o23s2q_flt.fits",
]
DEMO_IMAGE = "wfc3-ersii-g01-b6o-23-119.0-f140w_drz_sci.fits"

try:
    from .. import utils, model
except:
    from grizli import utils, model


class DrizzlePSF(object):
    epsf_obj = None
    driz_image = None

    def __init__(
        self,
        flt_files=DEMO_LIST,
        info=None,
        driz_image=DEMO_IMAGE,
        driz_hdu=None,
        beams=None,
        full_flt_weight=True,
        csv_file=None,
        epsf_obj=None,
    ):
        """
        Object for making drizzled PSFs

        Parameters
        ----------
        flt_files : list
            List of FLT files that were used to create the drizzled image.

        driz_image : str
            Filename of the drizzled image.

        """
        if info is None:
            if beams is not None:
                info = self._get_wcs_from_beams(beams)
            else:
                if flt_files is None:
                    info = self._get_wcs_from_hdrtab(driz_image)
                    if info is None:
                        info = self._get_wcs_from_csv(driz_image, csv_file=csv_file)
                else:
                    info = self._get_flt_wcs(flt_files, full_flt_weight=full_flt_weight)

        self.flt_keys, self.wcs, self.footprint = info
        self.flt_files = list(np.unique([key[0] for key in self.flt_keys]))

        if epsf_obj is None:
            self.epsf_obj = utils.EffectivePSF()
            self.epsf_obj.load_jwst_stdpsf()
        else:
            self.epsf_obj = epsf_obj

        if driz_hdu is None:
            self.driz_image = driz_image
            self.driz_header = pyfits.getheader(driz_image)
            self.driz_wcs = pywcs.WCS(self.driz_header)
            self.driz_pscale = utils.get_wcs_pscale(self.driz_wcs)
        else:
            self.driz_image = driz_image
            self.driz_header = driz_hdu.header
            self.driz_wcs = pywcs.WCS(self.driz_header)
            self.driz_pscale = utils.get_wcs_pscale(self.driz_wcs)

        self.driz_wcs.pscale = self.driz_pscale

    @property
    def ePSF(self):
        """
        For back compatibility, `ePSF` attribute renamed to `epsf_obj`
        """
        return self.epsf_obj

    @staticmethod
    def _get_flt_wcs(flt_files, full_flt_weight=True):
        """
        Get WCS info from FLT FITS files

        Parameters
        ----------
        flt_files : list
            List of filenames

        full_flt_weight : bool
            If True, set `expweight` from 'ERR' extension of the FITS files,
            else weight by EXPTIME keyword
        """
        from shapely.geometry import Polygon, Point

        from grizli import utils

        wcs = OrderedDict()
        footprint = OrderedDict()

        flt_keys = []
        for file in flt_files:
            flt_j = pyfits.open(file)
            for ext in range(1, 5):
                if ("SCI", ext) in flt_j:
                    key = file, ext
                    wcs[key] = pywcs.WCS(flt_j["SCI", ext], relax=True, fobj=flt_j)

                    wcs[key].pscale = utils.get_wcs_pscale(wcs[key])

                    if (("ERR", ext) in flt_j) & full_flt_weight:
                        wht = (1.0 / flt_j["ERR", ext].data ** 2).astype(np.float32)
                        wht[~np.isfinite(wht)] = 0
                        wcs[key].expweight = np.pad(wht, 32)
                    else:
                        wcs[key].expweight = flt_j[0].header["EXPTIME"]

                    footprint[key] = Polygon(wcs[key].calc_footprint())
                    flt_keys.append(key)

        return flt_keys, wcs, footprint

    @staticmethod
    def _get_wcs_from_beams(beams):
        """
        TBD
        """
        from shapely.geometry import Polygon, Point

        from grizli import utils

        wcs = OrderedDict()
        footprint = OrderedDict()

        flt_keys = []
        for beam in beams:
            # flt_j = pyfits.open(file)
            file = beam.direct.parent_file
            ext = beam.direct.sci_extn

            key = file, ext
            wcs[key] = beam.direct.wcs.copy()

            wcs[key].pscale = utils.get_wcs_pscale(wcs[key])

            wcs[key].expweight = beam.grism.exptime

            footprint[key] = Polygon(wcs[key].calc_footprint())
            flt_keys.append(key)

        return flt_keys, wcs, footprint

    @staticmethod
    def _get_wcs_from_csv(drz_file, csv_file=None):
        """
        Read the attached CSV file that contains exposure WCS info
        """
        import glob

        if csv_file is None:
            csv_file = drz_file.split("_drz_sci")[0].split("_drc_sci")[0]
            csv_file += "_wcs.csv"

            if not os.path.exists(csv_file):
                print(f"No WCS CSV file {csv_file} found for {drz_file}")
                return None
            else:
                print(f"Get exposure WCS from {csv_file}")

        csv = utils.read_catalog(csv_file, format="csv")
        flt_keys = []
        wcs = {}
        footprint = {}

        for row in csv:
            key = row["file"], row["ext"]
            _h = pyfits.Header()
            for c in row.colnames:
                _h[c] = row[c]

            flt_keys.append(key)
            wcs[key] = pywcs.WCS(_h, relax=True)

            wcs[key].pscale = utils.get_wcs_pscale(wcs[key])

            if "EXPTIME" in _h:
                wcs[key].expweight = _h["EXPTIME"]
            else:
                wcs[key].expweight = 1

            sr = utils.SRegion(wcs[key])
            footprint[key] = sr.shapely[0]

        return flt_keys, wcs, footprint

    @staticmethod
    def _get_wcs_from_hdrtab(drz_file):
        """
        Read tabulated exposure WCS info from the HDRTAB
        extension of an AstroDrizzle output file
        """
        from shapely.geometry import Polygon, Point

        drz = pyfits.open(drz_file)
        if "HDRTAB" not in drz:
            print("No HDRTAB extension found in {0}".format(drz_file))
            return None

        hdr = utils.GTable(drz["HDRTAB"].data)
        wcs = OrderedDict()
        footprint = OrderedDict()

        flt_keys = []
        N = len(hdr)

        if "CCDCHIP" in hdr.colnames:
            ext_key = "CCDCHIP"
        else:
            ext_key = "EXTNAME"

        for i in range(N):
            h = pyfits.Header()
            for c in hdr.colnames:
                try:
                    h[c] = hdr[c][i]
                except:
                    h[c] = 1

            key = (h["ROOTNAME"], h[ext_key])
            flt_keys.append(key)
            wcs[key] = pywcs.WCS(h, relax=True)
            wcs[key].pscale = utils.get_wcs_pscale(wcs[key])

            if "EXPTIME" in h:
                wcs[key].expweight = h["EXPTIME"]
            else:
                wcs[key].expweight = 1

            footprint[key] = Polygon(wcs[key].calc_footprint())

        return flt_keys, wcs, footprint

    def get_driz_cutout(
        self,
        ra=53.06967306,
        dec=-27.72333015,
        size=15,
        get_cutout=False,
        N=None,
        odd=True,
    ):
        """
        TBD
        """
        xy = self.driz_wcs.all_world2pix(np.array([[ra, dec]]), 0)[0]
        xyp = np.asarray(np.round(xy), dtype=int)
        if N is None:
            N = int(
                np.round(size * self.wcs[self.flt_keys[0]].pscale / self.driz_pscale)
            )

        slx = slice(xyp[0] - N, xyp[0] + N + odd)
        sly = slice(xyp[1] - N, xyp[1] + N + odd)

        wcs_slice = model.ImageData.get_slice_wcs(self.driz_wcs, slx, sly)

        wcs_slice.pscale = utils.get_wcs_pscale(wcs_slice)

        # outsci = np.zeros((2*N,2*N), dtype=np.float32)
        # outwht = np.zeros((2*N,2*N), dtype=np.float32)
        # outctx = np.zeros((2*N,2*N), dtype=np.int32)
        if get_cutout > 1:
            os.system(
                "getfits -o sub.fits {0} {1} {2} {3} {3}".format(
                    self.driz_image, xyp[0], xyp[1], 2 * N
                )
            )
            hdu = pyfits.open("sub.fits")
            return slx, sly, hdu
        elif get_cutout == 1:
            im = pyfits.open(self.driz_image)
            data = im[0].data[sly, slx] * 1
            header = utils.to_header(wcs_slice, relax=True)
            hdu = pyfits.PrimaryHDU(data=data, header=header)
            return slx, sly, pyfits.HDUList([hdu])

        return slx, sly, wcs_slice

    @staticmethod
    def _get_empty_driz(wcs):
        if hasattr(wcs, "pixel_shape"):
            sh = wcs.pixel_shape[::-1]
        else:
            if (not hasattr(wcs, "_naxis1")) & hasattr(wcs, "_naxis"):
                wcs._naxis1, wcs._naxis2 = wcs._naxis

            sh = (wcs._naxis2, wcs._naxis1)

        outsci = np.zeros(sh, dtype=np.float32)
        outwht = np.zeros(sh, dtype=np.float32)
        outctx = np.zeros(sh, dtype=np.int32)
        return outsci, outwht, outctx

    def go(self, ra=53.06967306, dec=-27.72333015):
        """
        Testing
        """
        import scipy.optimize

        wcs, footprint = None, None

        self = DrizzlePSF(
            info=(wcs, footprint), driz_image="cosmos-full-v1.2.8-f160w_drz_sci.fits"
        )

        slx, sly, wcs_slice = self.get_driz_cutout(ra=ra, dec=dec)
        xx, yy, drz_cutout = self.get_driz_cutout(ra=ra, dec=dec, get_cutout=True)

        psf = self.get_psf(ra=ra, dec=dec, filter="F160W", wcs_slice=wcs_slice)

        init = (0, 0, drz_cutout[0].data.sum())
        chi2 = self.objfun(init, self, ra, dec, wcs_slice, filter, drz_cutout)

        out = scipy.optimize.minimize(
            self.objfun,
            init,
            args=(self, ra, dec, wcs_slice, filter, drz_cutout),
            method="Powell",
            jac=None,
            hess=None,
            hessp=None,
            bounds=None,
            constraints=(),
            tol=1.0e-3,
            callback=None,
            options=None,
        )

        psf = self.get_psf(
            ra=ra + out.x[0] / 3600.0,
            dec=dec + out.x[1] / 3600.0,
            filter=filter,
            wcs_slice=wcs_slice,
            verbose=False,
        )

    @staticmethod
    def objfun_center(params, self, ra, dec, wcs_slice, filter, _Abg, sci, wht, ret):
        xoff, yoff = params[:2]

        psf_offset = self.get_psf(
            ra=ra + xoff / 3600.0 / 10.0,
            dec=dec + yoff / 3600.0 / 10.0,
            filter=filter,
            wcs_slice=wcs_slice,
            verbose=False,
        )

        _A = np.vstack([_Abg, psf_offset[1].data.flatten()])
        _Ax = _A * np.sqrt(wht.flatten())
        _y = (sci * np.sqrt(wht)).flatten()

        _res = np.linalg.lstsq(_Ax.T, _y, rcond=utils.LSTSQ_RCOND)

        model = _A.T.dot(_res[0]).reshape(sci.shape)
        chi2 = ((sci - model) ** 2 * wht).sum()
        print(params, chi2)
        if ret == 1:
            return params, _res, model, chi2
        else:
            return chi2

    @staticmethod
    def xobjfun_center(params, self, ra, dec, _Abg, sci, wht, ret):
        """
        Objective function for fitting the PSF model to a drizzled cutout
        """
        xoff, yoff = params[:2]

        filt = self.driz_header["FILTER"]

        psf_offset = self.get_psf(
            ra=ra + xoff / 3600.0 / 10.0,
            dec=dec + yoff / 3600.0 / 10.0,
            filter=filt.upper(),
            pixfrac=self.driz_header["PIXFRAC"],
            kernel=self.driz_header["KERNEL"],
            wcs_slice=self.driz_wcs,
            get_extended=filt.lower()[:2] in ["f1", "f0"],
            verbose=False,
        )

        if _Abg is not None:
            _A = np.vstack([_Abg, psf_offset[1].data.flatten()])
        else:
            _A = np.atleast_2d(psf_offset[1].data.flatten())

        _Ax = _A * np.sqrt(wht.flatten())
        _y = (sci * np.sqrt(wht)).flatten()

        _res = np.linalg.lstsq(_Ax.T, _y, rcond=utils.LSTSQ_RCOND)

        model = _A.T.dot(_res[0]).reshape(sci.shape)
        # chi2 = ((sci-model)**2*wht).sum()
        chi2 = _res[1][0]
        print(params, chi2)
        if ret == 1:
            return params, _res, model, chi2
        else:
            return chi2

    @staticmethod
    def objfun(params, self, ra, dec, wcs_slice, filter, drz_cutout):
        xoff, yoff = params[:2]
        psf = self.get_psf(
            ra=ra + xoff / 3600.0,
            dec=dec + yoff / 3600.0,
            filter=filter,
            wcs_slice=wcs_slice,
            verbose=False,
        )
        chi2 = ((psf[1].data * params[2] - drz_cutout[0].data) ** 2).sum()
        print(params, chi2)
        return chi2

    def get_psf(
        self,
        ra=53.06967306,
        dec=-27.72333015,
        filter="F140W",
        pixfrac=0.1,
        kernel="point",
        verbose=True,
        wcs_slice=None,
        get_extended=True,
        get_weight=False,
        ds9=None,
        npix=13,
        renormalize=True,
        xphase=0,
        yphase=0,
    ):
        """ """

        from drizzlepac import adrizzle
        from shapely.geometry import Polygon, Point

        pix = np.arange(-npix, npix + 1)

        # wcs_slice = self.get_driz_cutout(ra=ra, dec=dec)
        if wcs_slice is None:
            wcs_slice = self.driz_wcs.copy()

        outsci, outwht, outctx = self._get_empty_driz(wcs_slice)

        count = 1
        for key in self.flt_keys:
            if self.footprint[key].contains(Point(ra, dec)):
                file, ext = key
                if verbose:
                    print("{0}[SCI,{1}]".format(file, ext))

                xy = self.wcs[key].all_world2pix(np.array([[ra, dec]]), 0)[0]
                xyp = np.asarray(xy, dtype=int)  # np.round(xy))  # +1
                dx = xy[0] - int(xy[0]) + xphase
                dy = xy[1] - int(xy[1]) + yphase
                # dx = xy[0]-int(xy[0]) + 0.5
                # dy = xy[1]-int(xy[1]) + 0.5

                if ext == 2:
                    # UVIS
                    # print('UVIS1!')
                    chip_offset = 2051
                else:
                    chip_offset = 0

                psf_xy = self.epsf_obj.get_at_position(
                    xy[0], xy[1] + chip_offset, filter=filter
                )

                yp, xp = np.meshgrid(pix - dy, pix - dx, sparse=False, indexing="ij")
                if get_extended:
                    if filter in self.epsf_obj.extended_epsf:
                        extended_data = self.epsf_obj.extended_epsf[filter]
                    else:
                        extended_data = None
                else:
                    extended_data = None

                psf = self.epsf_obj.eval_ePSF(
                    psf_xy, xp, yp, extended_data=extended_data
                )

                flt_weight = self.wcs[key].expweight

                N = npix
                slx = slice(xyp[0] - N, xyp[0] + N + 1)
                sly = slice(xyp[1] - N, xyp[1] + N + 1)

                if hasattr(flt_weight, "ndim"):
                    if flt_weight.ndim == 2:
                        wslx = slice(xyp[0] - N + 32, xyp[0] + N + 1 + 32)
                        wsly = slice(xyp[1] - N + 32, xyp[1] + N + 1 + 32)
                        flt_weight = self.wcs[key].expweight[wsly, wslx]

                psf_wcs = model.ImageData.get_slice_wcs(self.wcs[key], slx, sly)

                psf_wcs.pscale = utils.get_wcs_pscale(self.wcs[key])

                try:
                    adrizzle.do_driz(
                        psf,
                        psf_wcs,
                        psf * 0 + flt_weight,
                        wcs_slice,
                        outsci,
                        outwht,
                        outctx,
                        1.0,
                        "cps",
                        1,
                        wcslin_pscale=1.0,
                        uniqid=1,
                        pixfrac=pixfrac,
                        kernel=kernel,
                        fillval=0,
                        stepsize=10,
                        wcsmap=None,
                    )
                except:
                    psf_wcs._naxis1, psf_wcs._naxis2 = psf_wcs._naxis

                    adrizzle.do_driz(
                        psf,
                        psf_wcs,
                        psf * 0 + flt_weight,
                        wcs_slice,
                        outsci,
                        outwht,
                        outctx,
                        1.0,
                        "cps",
                        1,
                        wcslin_pscale=1.0,
                        uniqid=1,
                        pixfrac=pixfrac,
                        kernel=kernel,
                        fillval=0,
                        stepsize=10,
                        wcsmap=None,
                    )

                if ds9 is not None:
                    count += 1
                    hdu = pyfits.HDUList(
                        [
                            pyfits.PrimaryHDU(),
                            pyfits.ImageHDU(
                                data=psf * 100, header=utils.to_header(psf_wcs)
                            ),
                        ]
                    )
                    ds9.set("frame {0}".format(count + 1))
                    ds9.set_pyfits(hdu)

        # ss = 1000000/2
        if renormalize:
            ss = 1.0 / outsci.sum() * psf.sum()
        else:
            ss = 1.0

        hdu = pyfits.HDUList(
            [
                pyfits.PrimaryHDU(),
                pyfits.ImageHDU(data=outsci * ss, header=utils.to_header(wcs_slice)),
            ]
        )

        if ds9 is not None:
            ds9.set("frame 2")
            ds9.set_pyfits(hdu)

        return hdu
